from torch import Tensor
import torch.nn as nn
import torch

from experiment_config import DatasetType


class ExperimentBaseModel(nn.Module):
    def __init__(self, dataset_type: DatasetType):
        super().__init__()
        self.dataset_type = dataset_type

    def forward(self, x) -> Tensor:
        raise NotImplementedError


class NiNBlock(nn.Module):
    def __init__(self, inplanes: int, planes: int) -> None:
        super().__init__()
        self.relu = nn.ReLU(inplace=True)

        self.conv1 = nn.Conv2d(inplanes, planes, kernel_size=3, stride=2, padding=1)
        self.bn1 = nn.BatchNorm2d(planes)

        self.conv2 = nn.Conv2d(planes, planes, kernel_size=1, stride=1)
        self.bn2 = nn.BatchNorm2d(planes)

        self.conv3 = nn.Conv2d(planes, planes, kernel_size=1, stride=1)
        self.bn3 = nn.BatchNorm2d(planes)

    def forward(self, x):
        x = self.conv1(x)
        x = self.bn1(x)
        x = self.relu(x)

        x = self.conv2(x)
        x = self.bn2(x)
        x = self.relu(x)

        x = self.conv3(x)
        x = self.bn3(x)
        x = self.relu(x)

        return x


class NiN(ExperimentBaseModel):
    def __init__(self, depth: int, width: int, base_width: int, dataset_type: DatasetType) -> None:
        super().__init__(dataset_type)

        self.base_width = base_width

        blocks = []
        blocks.append(NiNBlock(self.dataset_type.D[0], self.base_width * width))
        for _ in range(depth - 1):
            blocks.append(NiNBlock(self.base_width * width, self.base_width * width))
        self.blocks = nn.Sequential(*blocks)

        self.conv = nn.Conv2d(self.base_width * width, self.dataset_type.K, kernel_size=1, stride=1)
        self.bn = nn.BatchNorm2d(self.dataset_type.K)
        self.relu = nn.ReLU(inplace=True)

        self.avgpool = nn.AdaptiveAvgPool2d((1, 1))

    def forward(self, x):
        x = self.blocks(x)

        x = self.conv(x)
        x = self.bn(x)
        x = self.relu(x)

        x = self.avgpool(x)

        return x.squeeze()


# Notice that the effect of bias in conv layers will be cancelled by batchnorms,
# so for the most succinct implementation, we shouldn't use any bias here.
# However, in order that we can eliminate the influence of batchnorms, we set the bias to be true.
# Then, when calculating the complexity measures, reparametrise the network so that the batchnorms
# can be all set to N(0, 1).
def conv3x3(in_planes, out_planes, stride=1, groups=1, dilation=1):
    """3x3 convolution with padding"""
    return nn.Conv2d(in_planes, out_planes, kernel_size=3, stride=stride,
                     padding=dilation, groups=groups, bias=True, dilation=dilation)


def conv1x1(in_planes, out_planes, stride=1):
    """1x1 convolution"""
    return nn.Conv2d(in_planes, out_planes, kernel_size=1, stride=stride, bias=True)


class BasicBlock(nn.Module):
    expansion = 1

    def __init__(self, inplanes, planes, stride=1, downsample=None, groups=1,
                 base_width=64, dilation=1, norm_layer=None):
        super(BasicBlock, self).__init__()
        if norm_layer is None:
            norm_layer = nn.BatchNorm2d
        if groups != 1 or base_width != 64:
            raise ValueError('BasicBlock only supports groups=1 and base_width=64')
        if dilation > 1:
            raise NotImplementedError("Dilation > 1 not supported in BasicBlock")
        # Both self.conv1 and self.downsample layers downsample the input when stride != 1
        self.conv1 = conv3x3(inplanes, planes, stride)
        self.bn1 = norm_layer(planes)
        self.relu = nn.ReLU(inplace=True)
        self.conv2 = conv3x3(planes, planes)
        self.bn2 = norm_layer(planes)
        self.downsample = downsample
        self.stride = stride

    def forward(self, x):
        identity = x

        out = self.conv1(x)
        out = self.bn1(out)
        out = self.relu(out)

        out = self.conv2(out)
        out = self.bn2(out)

        if self.downsample is not None:
            identity = self.downsample(x)

        out += identity
        out = self.relu(out)

        return out


class CustomResNet(ExperimentBaseModel):
    """
    This implementation of ResNet is adapted from the torchvision model zoo.
    The original implementation can be found at:
    https://github.com/pytorch/vision/blob/b266c2f1a5c10f5caf22f5aea7418acc392a5075/torchvision/models/resnet.py
    """
    def __init__(self, depth: int, width: int, base_width: int, dataset_type: DatasetType) -> None:
        super().__init__(dataset_type)
        self.depth = depth
        self.width = width
        # self.base_width = base_width
        self.base_width = 64

        self.inplanes = 64
        self.dilation = 1
        replace_stride_with_dilation = [False, False, False]
        self.groups = 1
        num_classes = 10
        self._norm_layer = norm_layer = nn.BatchNorm2d
        block = BasicBlock

        if self.depth == 2:
            num_layers = [2, 2, 2, 2]
        elif self.depth == 3:
            num_layers = [3, 3, 3, 3]
        elif self.depth == 4:
            num_layers = [3, 4, 6, 3]
        elif self.depth == 5:
            num_layers = [3, 4, 10, 3]
        elif self.depth == 6:
            num_layers = [3, 4, 14, 3]
        else:
            raise NotImplementedError
        self.conv1 = nn.Conv2d(3, self.inplanes, kernel_size=7, stride=2, padding=3,
                               bias=True)
        self.bn1 = norm_layer(self.inplanes)
        self.relu = nn.ReLU(inplace=True)
        self.maxpool = nn.MaxPool2d(kernel_size=3, stride=2, padding=1)
        self.layer1 = self._make_layer(block, 64, num_layers[0], stride=1)
        self.layer2 = self._make_layer(block, 128, num_layers[1], stride=2)
        self.layer3 = self._make_layer(block, 256, num_layers[2], stride=2)
        self.layer4 = self._make_layer(block, 512, num_layers[3], stride=2)
        self.avgpool = nn.AdaptiveAvgPool2d((1, 1))
        self.fc = nn.Linear(512 * block.expansion, num_classes)

        for m in self.modules():
            if isinstance(m, nn.Conv2d):
                nn.init.kaiming_normal_(m.weight, mode='fan_out', nonlinearity='relu')
            elif isinstance(m, (nn.BatchNorm2d, nn.GroupNorm)):
                nn.init.constant_(m.weight, 1)
                nn.init.constant_(m.bias, 0)

    def _make_layer(self, block, planes, blocks, stride=1, dilate=False):
        norm_layer = self._norm_layer
        downsample = None
        previous_dilation = self.dilation
        if dilate:
            self.dilation *= stride
            stride = 1
        if stride != 1 or self.inplanes != planes * block.expansion:
            downsample = nn.Sequential(
                conv1x1(self.inplanes, planes * block.expansion, stride),
                norm_layer(planes * block.expansion),
            )

        layers = []
        layers.append(block(self.inplanes, planes, stride, downsample, self.groups,
                            self.base_width, previous_dilation, norm_layer))
        self.inplanes = planes * block.expansion
        for _ in range(1, blocks):
            layers.append(block(self.inplanes, planes, groups=self.groups,
                                base_width=self.base_width, dilation=self.dilation,
                                norm_layer=norm_layer))

        return nn.Sequential(*layers)

    def _forward_impl(self, x):
        # See note [TorchScript super()]
        x = self.conv1(x)
        x = self.bn1(x)
        x = self.relu(x)
        x = self.maxpool(x)

        x = self.layer1(x)
        x = self.layer2(x)
        x = self.layer3(x)
        x = self.layer4(x)

        x = self.avgpool(x)
        x = torch.flatten(x, 1)
        x = self.fc(x)

        return x

    def forward(self, x):
        return self._forward_impl(x)


# This is taken from https://github.com/akamaster/pytorch_resnet_cifar10/blob/master/resnet.py
import torch.nn.init as init
import torch.nn.functional as F

def _weights_init(m):
    classname = m.__class__.__name__
    #print(classname)
    if isinstance(m, nn.Linear) or isinstance(m, nn.Conv2d):
        init.kaiming_normal_(m.weight)

class LambdaLayer(nn.Module):
    def __init__(self, lambd):
        super(LambdaLayer, self).__init__()
        self.lambd = lambd

    def forward(self, x):
        return self.lambd(x)


class ProperBasicBlock(nn.Module):
    expansion = 1

    def __init__(self, in_planes, planes, stride=1, option='A'):
        super(ProperBasicBlock, self).__init__()
        self.conv1 = nn.Conv2d(in_planes, planes, kernel_size=3, stride=stride, padding=1)
        self.bn1 = nn.BatchNorm2d(planes)
        self.conv2 = nn.Conv2d(planes, planes, kernel_size=3, stride=1, padding=1)
        self.bn2 = nn.BatchNorm2d(planes)

        self.shortcut = nn.Sequential()
        if stride != 1 or in_planes != planes:
            if option == 'A':
                """
                For CIFAR10 ResNet paper uses option A.
                """
                self.shortcut = LambdaLayer(lambda x:
                                            F.pad(x[:, :, ::2, ::2], (0, 0, 0, 0, planes//4, planes//4), "constant", 0))
            elif option == 'B':
                self.shortcut = nn.Sequential(
                     nn.Conv2d(in_planes, self.expansion * planes, kernel_size=1, stride=stride),
                     nn.BatchNorm2d(self.expansion * planes)
                )

    def forward(self, x):
        out = F.relu(self.bn1(self.conv1(x)))
        out = self.bn2(self.conv2(out))
        out += self.shortcut(x)
        out = F.relu(out)
        return out


class SkipInitBlock(nn.Module):
    expansion = 1

    def __init__(self, in_planes, planes, stride=1):
        super(SkipInitBlock, self).__init__()
        self.conv1 = nn.Conv2d(in_planes, planes, kernel_size=3, stride=stride, padding=1)
        self.conv2 = nn.Conv2d(planes, planes, kernel_size=3, stride=1, padding=1)
        self.shortcut = nn.Sequential()
        if stride != 1 or in_planes != planes:
            self.shortcut = LambdaLayer(lambda x:
                                        F.pad(x[:, :, ::2, ::2], (0, 0, 0, 0, planes//4, planes//4), "constant", 0))
        self.skipinit_constant = nn.Parameter(torch.zeros(1, requires_grad=True))

    def forward(self, x):
        out = self.conv1(F.relu(x))
        out = self.conv2(F.relu(out))
        out = self.skipinit_constant * out
        out += self.shortcut(x)
        return out


class ProperResNet(ExperimentBaseModel):
    def __init__(self, block, num_blocks, dataset_type,
                 width=8, skipinit=False,
                 num_classes=10):
        super(ProperResNet, self).__init__(dataset_type)
        self.in_planes = 2 * width
        self.skipinit = skipinit

        self.conv1 = nn.Conv2d(3, self.in_planes, kernel_size=3, stride=1, padding=1)
        if not skipinit:
            self.bn1 = nn.BatchNorm2d(self.in_planes)
        self.layer1 = self._make_layer(block, self.in_planes, num_blocks[0], stride=1)
        self.layer2 = self._make_layer(block, self.in_planes*2, num_blocks[1], stride=2)
        self.layer3 = self._make_layer(block, self.in_planes*2, num_blocks[2], stride=2)
        self.linear = nn.Linear(self.in_planes, num_classes)

        self.apply(_weights_init)

    def _make_layer(self, block, planes, num_blocks, stride):
        strides = [stride] + [1] * (num_blocks - 1)
        layers = []
        for stride in strides:
            layers.append(block(self.in_planes, planes, stride))
            self.in_planes = planes * block.expansion

        return nn.Sequential(*layers)

    def forward(self, x):
        if self.skipinit:
            out = F.relu(self.conv1(x))
        else:
            out = F.relu(self.bn1(self.conv1(x)))
        out = self.layer1(out)
        out = self.layer2(out)
        out = self.layer3(out)
        out = F.avg_pool2d(out, out.size()[3])
        out = out.view(out.size(0), -1)
        out = self.linear(out)
        return out


def p_resnet_initialiser(depth: int, width: int, base_width: int, dataset_type: DatasetType):
    if depth == 2:
        return ProperResNet(ProperBasicBlock, [3, 3, 3], dataset_type, width=width)
    elif depth == 3:
        return ProperResNet(ProperBasicBlock, [5, 5, 5], dataset_type, width=width)
    elif depth == 4:
        return ProperResNet(ProperBasicBlock, [6, 6, 6], dataset_type, width=width)
    elif depth == 5:
        return ProperResNet(ProperBasicBlock, [8, 8, 8], dataset_type, width=width)
    elif depth == 6:
        return ProperResNet(ProperBasicBlock, [9, 9, 9], dataset_type, width=width)
    else:
        raise NotImplementedError


def skipinit_resnet_initialiser(depth: int, width: int, dataset_type: DatasetType):
    return ProperResNet(SkipInitBlock, [depth, depth, depth], dataset_type, width=width, skipinit=True)


# Constants
IMAGE_WIDTH = 32
IMAGE_HEIGHT = 32
COLOR_CHANNELS = 3
N_CLASSES = 10


class MLP(ExperimentBaseModel):
    def __init__(self, n_hidden_nodes, n_hidden_layers, dataset_type):
        super(MLP, self).__init__(dataset_type)
        self.n_hidden_nodes = n_hidden_nodes
        self.n_hidden_layers = n_hidden_layers
        # Set up perceptron layers and add dropout
        self.fc1 = torch.nn.Linear(IMAGE_WIDTH * IMAGE_WIDTH * COLOR_CHANNELS,
                                   n_hidden_nodes)

        self.hidden_layers = nn.ModuleList([
            nn.Linear(n_hidden_nodes, n_hidden_nodes) for _ in range(self.n_hidden_layers)
        ])

        self.out = torch.nn.Linear(n_hidden_nodes, N_CLASSES)

    def forward(self, x):
        x = x.view(-1, IMAGE_WIDTH * IMAGE_WIDTH * COLOR_CHANNELS)
        x = nn.functional.relu(self.fc1(x))

        for i in range(self.n_hidden_layers):
            x = self.hidden_layers[i](x)
            x = nn.functional.relu(x)
        return self.out(x)


def mlp_initialiser(depth: int, width: int, dataset_type: DatasetType):
    return MLP(n_hidden_nodes=width*8, n_hidden_layers=depth, dataset_type=dataset_type)


class VanillaCNN(ExperimentBaseModel):
    def __init__(self, n_hidden_nodes, n_hidden_layers, dataset_type):
        super(VanillaCNN, self).__init__(dataset_type)
        self.n_hidden_nodes = n_hidden_nodes
        self.n_hidden_layers = n_hidden_layers

        self.features = nn.Sequential(
            nn.Conv2d(3, 64, kernel_size=11, stride=4, padding=5),
            nn.ReLU(inplace=True),
            nn.MaxPool2d(kernel_size=2, stride=2),
            nn.Conv2d(64, 128, kernel_size=5, padding=2),
            nn.ReLU(inplace=True),
            nn.MaxPool2d(kernel_size=2, stride=2),
            nn.Conv2d(128, self.n_hidden_nodes, kernel_size=3, padding=1),
            nn.ReLU(inplace=True),
        )
        self.hidden_layers = nn.ModuleList(
            [
                nn.Sequential(
                    nn.Conv2d(self.n_hidden_nodes, self.n_hidden_nodes, kernel_size=3, padding=1),
                    nn.ReLU(inplace=True),
                ) for _ in range(self.n_hidden_layers - 1)
            ]
        )
        self.final_pool = nn.MaxPool2d(kernel_size=2, stride=2)
        self.fc = nn.Linear(self.n_hidden_nodes, 10)

    def forward(self, x):
        x = self.features(x)
        for i in range(self.n_hidden_layers-1):
            x = self.hidden_layers[i](x)
        x = self.final_pool(x)
        return self.fc(x.view(x.size(0), -1))


def cnn_initialiser(depth: int, width: int, dataset_type: DatasetType):
    return VanillaCNN(n_hidden_nodes=width*8, n_hidden_layers=depth, dataset_type=dataset_type)
